////////////////////////////////////////////////////////////
// strips the parameters from an url
// and returns the "base" url
function getBaseURL(urlstring)
{
    var str = String(urlstring);

    var iQuestionMark = str.indexOf("?");

    if( iQuestionMark == -1 )
    {
        return str;
    }
    else
    {
        return str.substr(0, iQuestionMark);
    }
}

////////////////////////////////////////////////////////////
// urlencodes a string
// example: "Hello World" -> "Hello%20World"
function urlencode( str )
{
    return encodeURIComponent( str );
}

////////////////////////////////////////////////////////////
// urldecodes a string
// example: "Hello%20World" -> "Hello World"
function urldecode( str )
{
    var sReturn = "";

    //convert all + to %20
    str = str.replace( /\+/g, "%20" );

    // well, with some encoded strings using PHPs urlencode
    // decodeURIComponent throws an error
    // if so, use the standard unescape function
    try {
        sReturn = decodeURIComponent( str );
    } catch( ex ) {
        sReturn = unescape( str );
    }

    return sReturn;
}

////////////////////////////////////////////////////////////
// tests a value if it it contains a url
function isUrl( sValue ) {
      var regexp = /(ftp|http|https):\/\/(\w+:{0,1}\w*@)?(\S+)(:[0-9]+)?(\/|\/([\w#!:.?+=&%@!\-\/]))?/
      return regexp.test(sValue);
}

////////////////////////////////////////////////////////////
// array search functions
function searchParameterArray( sParameterName, aArray ) {
    for( var sKey in aArray ) {
        var aSub = aArray[sKey];

        if( aSub[0] == sParameterName ) {
            return aSub;
        }
    }

    return false;
}

function removeDuplsFromArray( aArray ) {

    var aNew = new Array();

    for( var sKey in aArray ) {
        var ukn = searchParameterArray( aArray[sKey][0], aNew );

        if( typeof(ukn) == "boolean" ) {
            aNew.push( aArray[sKey] );
        }
    }

    return aNew;
}

////////////////////////////////////////////////////////////
// well, to get the blur event of the empty label textbox
// we can't (for whatever reason) use addEventlistener
// so i attached the event via setAttribute("onblur"...);
//
// this function dispatches the event back to the UrlParamsObj
// instance, named oUrlParams (see sidebar.xul)
//
function dispatchEventToUrlParams(event) {
    if( oUrlParams ) {
        oUrlParams.debugDump("dispatchEventToUrlParams: dispatching.");
        oUrlParams.handleEvent(event);
    }
}

////////////////////////////////////////////////////////////
// the Object
function UrlParamsObj(theWindow)
{
    this.construct(theWindow);
}

UrlParamsObj.prototype = {
    oBrowserWindow : null,
    oTabBrowser    : null,
    oPrefsService  : null,

    selectedBrowser: null,
    theWindow:       null,

    bUrlDeEncodeUrl              : true,
    bDebugMode                   : false,    //if true: debug messages are printed to the console, start firefox with -console to see them
    bParseUrlsInValues           : false,
    sAutocomplete                : 'off',
    bAutocompleteShowArrowButton : false,
    bKeepUncheckedParams         : false,
    bSortParameters              : true,

    aUncheckedParameters         : null,

    debugDump: function(sText)
    {
        var consoleService = Components.classes["@mozilla.org/consoleservice;1"]
                     .getService(Components.interfaces.nsIConsoleService);
        consoleService.logStringMessage("[urlparams] " + sText);
    },

    construct: function(theWindow)
    {
        this.debugDump("constructing UrlParamsObj...");

        this.theWindow = theWindow;

        const kWindowMediatorContractID = "@mozilla.org/appshell/window-mediator;1";
        const kWindowMediatorIID = Components.interfaces.nsIWindowMediator;
        const kWindowMediator = Components.classes[kWindowMediatorContractID].getService(kWindowMediatorIID);

        this.oBrowserWindow = kWindowMediator.getMostRecentWindow("navigator:browser");
        this.oTabBrowser    = this.oBrowserWindow.getBrowser();

        this.getSettings();

        this.theWindow.addEventListener("load",   this, false);
        this.theWindow.addEventListener("unload", this, false);
    },

    init: function(event)
    {
        this.debugDump("initializing UrlParamsObj...");

        this.theWindow.addEventListener("dblclick", this, false);
        this.initDragDrop();

        this.oTabBrowser.mTabBox.addEventListener('select', this, true);

        this.changeLoadEvent();

        var button = this.theWindow.document.getElementById("get-button");
        button.addEventListener("command", this, true);

        button = this.theWindow.document.getElementById("get-button-newtab");
        button.addEventListener("command", this, true);

        button = this.theWindow.document.getElementById("clear-checkboxes");
        button.addEventListener("command", this, true);

        button = this.theWindow.document.getElementById("get-url");
        button.addEventListener("command", this, true);

        var img = this.theWindow.document.getElementById("img-plusget");
        img.addEventListener("click", this, true);

        var img = this.theWindow.document.getElementById("img-pluspost");
        img.addEventListener("click", this, true);

        //load current pages url
        this.onTabFocus(null);
    },

    getSettings: function()
    {
        if( this.oPrefsService == null )
        {
            this.oPrefsService = Components.classes["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefBranch);
        }

        try
        {
            this.bDebugMode                   = this.oPrefsService.getBoolPref("urlparams.debug");
            this.bUrlDeEncodeUrl              = this.oPrefsService.getBoolPref("urlparams.deencodeurl");
            this.bParseUrlsInValues           = this.oPrefsService.getBoolPref("urlparams.parseurlsinvalues");
            this.sAutocomplete                = this.oPrefsService.getCharPref("urlparams.autocomplete");
            this.bAutocompleteShowArrowButton = this.oPrefsService.getBoolPref("urlparams.autocompletearrowbutton");
            this.bKeepUncheckedParams         = this.oPrefsService.getBoolPref("urlparams.keepuncheckedparams");
            this.bSortParameters              = this.oPrefsService.getBoolPref("urlparams.sortparameters");
        }
        catch( exception )
        {}
    },


    removeGridChilds: function()
    {
        //get rows objects
        var oGridRowsGet  = this.theWindow.document.getElementById("gridrows-get");
        var oGridRowsPost = this.theWindow.document.getElementById("gridrows-post");

        //del childs
        var iRows = oGridRowsGet.childNodes.length;
        for(var i = 0; i < iRows; i++)
        {
            oGridRowsGet.removeChild(oGridRowsGet.childNodes[0]);
        }

        var iRows = oGridRowsPost.childNodes.length;
        for(var i = 0; i < iRows; i++)
        {
            oGridRowsPost.removeChild(oGridRowsPost.childNodes[0]);
        }
    },

    getAutocompleteTextboxSearchParam: function( sAutocompleteSearchParam, bDoNotAlterParam ) {

        switch( this.sAutocomplete ) {
            default:
            case 'off':
                // return an empty string to prevent autocomplete from working
                return "";
                break;

            case 'own':
                // if we should use our own history,
                // we just prepend a string to each searchparam
                // but in fact, we are using the global history
                if( !bDoNotAlterParam ) {
                    sAutocompleteSearchParam = "urlparams_autocomplete_" + sAutocompleteSearchParam;
                }

            case 'global':
                return sAutocompleteSearchParam;
                break;
        }
    },

    setAutocompleteTextboxAttributes: function( oTextbox, bRemoveAttribs, sAutocompleteSearchParam, bDoNotAlterParam ) {
        if( this.sAutocomplete == "off" ) {
            return;
        }

        if( bRemoveAttribs ) {
            oTextbox.setAttribute("type"                   , "");
            oTextbox.setAttribute("autocompletesearch"     , "");
            oTextbox.setAttribute("autocompletesearchparam", "");
            oTextbox.setAttribute("completeselectedindex"  , "");
            oTextbox.setAttribute("enablehistory"          , "");
        } else {
            sSearchValue = this.getAutocompleteTextboxSearchParam(sAutocompleteSearchParam,bDoNotAlterParam);

            oTextbox.setAttribute("type"                   , "autocomplete");
            oTextbox.setAttribute("autocompletesearch"     , "form-history");
            oTextbox.setAttribute("autocompletesearchparam", sSearchValue);
            oTextbox.setAttribute("completeselectedindex"  , "true");
            if( this.bAutocompleteShowArrowButton ) {
                oTextbox.setAttribute("enablehistory"          , "true");
            }
        }
    },

    /*
    createAutocompleteTextbox: function( sAutocompleteSearchParam, bDoNotAlterParam ) {

        if( this.theWindow && this.theWindow.document ) {

            var oTextbox = this.theWindow.document.createElement("textbox");

            var sSearchValue = this.getAutocompleteTextboxSearchParam(sAutocompleteSearchParam, bDoNotAlterParam);

            switch( this.sAutocomplete ) {
                default:
                case 'off':
                    // do nothing, just create a textbox
                    break;

                case 'own':
                case 'global':
                    oTextbox.setAttribute("type"                   , "autocomplete");
                    oTextbox.setAttribute("autocompletesearch"     , "form-history");
                    oTextbox.setAttribute("autocompletesearchparam", sSearchValue);
                    oTextbox.setAttribute("completeselectedindex"  , "true");
                    if( this.bAutocompleteShowArrowButton ) {
                        oTextbox.setAttribute("enablehistory"          , "true");
                    }
                    break;
            }

            return oTextbox;

        } else {
            return null;
        }
    },
    */

    changeEmptyValueInput: function( oEmptyLabelInput ) {

        if( this.sAutocomplete == 'off' ) {
            this.debugDump("changeEmptyValueInput: autocomplete disabled. do nothing.");
            return;
        }

        if( oEmptyLabelInput && oEmptyLabelInput.theValueTextbox ) {
            if( oEmptyLabelInput.value != "" ) {

                //get the autocompletesearchparam (depending on the autocomplete setting)
                var sSearchValue = this.getAutocompleteTextboxSearchParam(oEmptyLabelInput.value, false);

                if( (oEmptyLabelInput.theValueTextbox.getAttribute("type") == "autocomplete") && (oEmptyLabelInput.theValueTextbox.getAttribute("autocompletesearchparam") == sSearchValue) ) {
                    this.debugDump("changeEmptyValueInput: already autocomplete and equal searchparam. do nothing");
                } else if( oEmptyLabelInput.theValueTextbox.getAttribute("multiline") == "true" ) {
                    this.debugDump("changeEmptyValueInput: value textbox is multiline. do nothing");
                } else {
                    this.debugDump("changeEmptyValueInput: converting textbox into autocomplete textbox");

                    //saving value, cause this gets lost during convertion
                    var sValue = oEmptyLabelInput.theValueTextbox.value;

                    //first, restore the textbox into a normal one, then
                    //convert the textbox into an autocomplete textbox
                    this.setAutocompleteTextboxAttributes(oEmptyLabelInput.theValueTextbox, true);
                    this.setAutocompleteTextboxAttributes(oEmptyLabelInput.theValueTextbox, false, oEmptyLabelInput.value, false);

                    oEmptyLabelInput.theValueTextbox.value = sValue;
                }
            } else {
                this.debugDump("changeEmptyValueInput: converting autocomplete textbox into normal textbox");
                var sValue = oEmptyLabelInput.theValueTextbox.value;

                //convert the autocomplete textbox into a normal one
                this.setAutocompleteTextboxAttributes(oEmptyLabelInput.theValueTextbox, true);

                oEmptyLabelInput.theValueTextbox.value = sValue;
            }
        }
    },

    addEmptyRow: function(id, bMultiline)
    {
        //get rows objects
        var oGridRows = this.theWindow.document.getElementById(id);

        //append check/textboxes for new param
        var row      = this.theWindow.document.createElement("row");
        var checkbox = this.theWindow.document.createElement("checkbox");

        // create the value textbox
        var textbox  = this.theWindow.document.createElement("textbox");
        // if bMultiline make a multiline textbox out of it
        if( bMultiline ) {
            textbox.setAttribute("multiline", "true");
        }

        // create the autocomplete label textbox
        var label    = this.theWindow.document.createElement("textbox");
        // set second parameter to true, because the empty label should give the same results
        // regardless if we should use global or our own autocomplete history
        this.setAutocompleteTextboxAttributes( label, false, "urlparams_autocomplete_emptyParamLabel", true);

        checkbox.setAttribute("checked", "true");

        label.setAttribute("width", "60");
        label.setAttribute("flex" , "1");

        //well, append the textbox as attribute to the label
        //so i don't need to generate unique ids
        //we only need this, if autocomplete is enabled (see updateFormHistory)
        label.theValueTextbox = textbox;

        row.appendChild(checkbox);
        row.appendChild(label);
        row.appendChild(textbox);
        oGridRows.appendChild(row);

        // add event listeners after the dynamically created elemented is added to the DOM
        // otherwise the addEventListener has no effect
        if( this.sAutocomplete != 'off' ) {
            label.addEventListener("popuphidden", this, true);
            label.addEventListener("blur", this, true);
        }
    },

    addEmptyRows: function()
    {
        this.addEmptyRow("gridrows-get");
        this.addEmptyRow("gridrows-post");
    },

    addRow: function(id, sName, sValue, bChecked)
    {
        if( arguments.length < 4 ) {
            var bChecked = true;
        }

        //get rows objects
        var oGridRows = this.theWindow.document.getElementById(id);

        var row      = this.theWindow.document.createElement("row");
        var label    = this.theWindow.document.createElement("label");
        var checkbox = this.theWindow.document.createElement("checkbox");
        var textbox  = this.theWindow.document.createElement("textbox");

        // causes the child nodes to be centered on a horizontal line
        row.setAttribute("align", "center");

        checkbox.setAttribute("checked" , (bChecked ? "true" : "false") );
        checkbox.setAttribute("disabled", "false");

        label.setAttribute("value", sName);

        // if sValue contains at least one \n
        // create a multiline textbox
        //
        // and only create a autocomplete box, if the content is not multiline!
        if( sValue.match(/\n/) )
        {
            textbox.setAttribute("multiline", "true");
        } else {
            this.setAutocompleteTextboxAttributes(textbox, false, sName, false);
        }

        textbox.setAttribute("value", sValue);
        textbox.setAttribute("tooltiptext", sValue);

        row.appendChild(checkbox);
        row.appendChild(label);
        row.appendChild(textbox);

        oGridRows.appendChild(row);

        // event listener for updating the tooltiptext
        textbox.addEventListener("change", this, true);
    },

    fill: function()
    {
        var urlfield     = this.theWindow.document.getElementById("url_wo_params");
        var refurlfield  = this.theWindow.document.getElementById("url_referer");

        var boxGetLabel  = this.theWindow.document.getElementById("groupboxGetLabel");
        var boxPostLabel = this.theWindow.document.getElementById("groupboxPostLabel");

        //get the most recent entry in the session history
        var sessionHistory = this.selectedBrowser.webNavigation.sessionHistory;

        this.debugDump("fill: sessionHistory.index = " + sessionHistory.index);

        if( sessionHistory.index == -1 )
        {
            urlfield.value    = "";
            refurlfield.value = "";

            boxGetLabel.value  = "GET (0)";
            boxPostLabel.value = "POST (0)";

            this.addEmptyRows();
            return;
        }

        var entry = sessionHistory.getEntryAtIndex(sessionHistory.index, 0);
        entry = entry.QueryInterface(Components.interfaces.nsISHEntry);

        //create a url object to manipulate the uri with
        var url =
        Components.classes["@mozilla.org/network/standard-url;1"].
        createInstance(Components.interfaces.nsIStandardURL);
        url = url.QueryInterface(Components.interfaces.nsIURL);

        //grab the uri from the session history entry and extract the query from the uri
        url.spec = entry.URI.spec;
        var query = url.query;
        url.query = "";
        var uri = url.spec;

        //grab the referring uri from the session history entry
        var referringUri = entry.referrerURI;
        if (referringUri)
        {
            referringUri = referringUri.QueryInterface(Components.interfaces.nsIURI);
            referringUri = referringUri.spec;
        }
        else
        {
            referringUri = "";
        }

        //update the uri and referringUri textboxes
        urlfield.value    = uri;
        refurlfield.value = referringUri;

        //parse the query string and fill the get box
        var iGets = this.fillBox("gridrows-get", query);
        boxGetLabel.value = "GET ("+iGets+")";

        //handle the post data
        var iPosts = 0;
        if (entry.postData)
        {
            this.debugDump("fill: entry has post data!");
            this.debugDump("fill: here they are:");

            //catch the following operations
            //in the case of a file upload, FF doesn't store the postdata in the history,
            //but "pretends" that we have postdata
            try {
                var postDataStream = entry.postData;
                postDataStream.QueryInterface(Components.interfaces.nsISeekableStream).seek(0, 0);

                //create an input stream for reading the post data from
                var inputStream = Components.classes["@mozilla.org/scriptableinputstream;1"].createInstance(Components.interfaces.nsIScriptableInputStream);
                inputStream.init(entry.postData);

                var postData = inputStream.read(0xFFFFFFFF);

                this.debugDump("----- raw begin -----");
                this.debugDump(postData);
                this.debugDump("----- raw end -----");

                ////split headers and actual post data
                //var aPostData = postData.split("\r\n\r\n");
                //
                ////examine headers
                //var bMultipart = false;
                //var sBoundary  = "";
                //var aHeaders   = aPostData[0].split("\r\n");
                //
                //var rxUrl   = /\[url\]$/;
                //
                //for( var iHeaders in aHeaders ) {
                //    var sHeader = aHeaders[iHeaders].toLowerCase();
                //
                //    if( sHeader.
                //}

                var sParams = postData.split("\r\n\r\n")[1];

                // parse the data and fill the post box
                iPosts = this.fillBox("gridrows-post", sParams);
            } catch( e ) {
                // do nothing
                this.debugDump("an error occured, maybe there was actual no postdata. msg: " + (e ? e : ''));
            }
        }
        boxPostLabel.value = "POST ("+iPosts+")";

        this.addEmptyRows();
    },

    fillBoxFromArray: function( id, dataArray ) {

        this.debugDump("fillBoxFromArray...");
        this.debugDump("id: " + id);
        this.debugDump("dataArray.length: " + dataArray.length);

        var iParams = 0;

        //create new childs
        for(var sKey in dataArray)
        {
            var name     = dataArray[sKey][0];
            var value    = dataArray[sKey][1];
            var bChecked = dataArray[sKey][2];

            if( typeof(bChecked) == "undefined" ) {
                bChecked = true;
            }

            if( value != "" ) {
                if( this.bUrlDeEncodeUrl ) {
                    value = urldecode( value );
                }
            }

            if( name != "" )
            {
                iParams++;

                if( this.bUrlDeEncodeUrl ) {
                    name = urldecode( name );
                }

                // check if value contains a url with parameters
                var sValue = String(value);
                if( this.bParseUrlsInValues && (sValue.search(/\?/) != -1) ) {

                    this.debugDump( "parse urls in values:" );

                    var aUrlParts = sValue.split('?');

                    this.debugDump( "parts: [" + aUrlParts[0] + "][" + aUrlParts[1] + "]"  );

                    this.addRow(id, name+"[url]", aUrlParts[0]);

                    if( aUrlParts[1] ) {
                        var aUrlPartParams = aUrlParts[1].split('&');

                        for(var ip = 0; ip < aUrlPartParams.length; ip++) {
                            this.addRow(id, name+"[var]", aUrlPartParams[ip], bChecked);
                        }
                    }

                } else {
                    this.addRow(id, name, value, bChecked);
                }
            }
        }

        this.debugDump("fillBoxFromArray... done (" + iParams + ")");

        return iParams;
    },

    fillBox: function( id, dataString )
    {
        this.debugDump("fillBox...");
        this.debugDump("id: " + id);
        this.debugDump("dataString: " + dataString);

        var iParams = 0;

        if( dataString ) {
            var aParams = dataString.split("&");

            if( this.bSortParameters )
            {
                aParams.sort();
            }

            //create new childs
            for(var i = 0; i < aParams.length; i++)
            {
                //var pair  = aParams[i].split("=");
                //var name  = pair[0];
                //var value = pair[1];
                //
                //if( typeof(pair[1]) == "undefined" )
                //{
                //    value = "";
                //}

                var iSplitAt = aParams[i].indexOf('=');
                var name     = "";
                var value    = "";

                if( iSplitAt != -1 ) {
                    name  = aParams[i].substring(0,iSplitAt);
                    value = aParams[i].substring(iSplitAt+1);
                } else {
                    name  = aParams[i];
                }

                if( value != "" ) {
                    if( this.bUrlDeEncodeUrl ) {
                        value = urldecode( value );
                    }
                }

                if( name != "" )
                {
                    iParams++;

                    if( this.bUrlDeEncodeUrl ) {
                        name = urldecode( name );
                    }

                    // check if value contains a url with parameters
                    var sValue = String(value);
                    if( this.bParseUrlsInValues && (sValue.search(/\?/) != -1) ) {

                        var aUrlParts = sValue.split('?');

                        this.addRow(id, name+"[url]", aUrlParts[0]);

                        if( aUrlParts[1] ) {
                            var aUrlPartParams = aUrlParts[1].split('&');

                            for(var ip = 0; ip < aUrlPartParams.length; ip++) {
                                this.addRow(id, name+"[var]", aUrlPartParams[ip]);
                            }
                        }

                    } else {
                        this.addRow(id, name, value);
                    }
                }
            }
        }

        if( this.bKeepUncheckedParams == true && this.aUncheckedParameters != null) {

            this.debugDump("Creating unchecked parameters...");

            if( this.aUncheckedParameters[id] ) {
                this.fillBoxFromArray(id, this.aUncheckedParameters[id]);

                //clear unchecked parameters
                this.aUncheckedParameters[id] = new Array();
            }
        }

        this.debugDump("fillBox... done (" + iParams + ")");

        return iParams;
    },

    getDataArray: function(id, bIgnoreCheck) {

        if( arguments.length < 2 ) {
            var bIgnoreCheck = false;
        }

        var aParams = new Array();
        var rxUrl   = /\[url\]$/;
        var rxVar   = /\[var\]$/;

        //get rows object
        var oGridRows = this.theWindow.document.getElementById(id);

        //check if their are childs (their is minimum 1 (the empty row), so also check
        //if this is the first element...
        if( (oGridRows.childNodes.length > 0) && (oGridRows.childNodes[0].childNodes[1].value != "") )
        {
            for(var i = 0; i < oGridRows.childNodes.length; i++)
            {
                var oCheck = oGridRows.childNodes[i].childNodes[0];
                var oLabel = oGridRows.childNodes[i].childNodes[1];
                var oValue = oGridRows.childNodes[i].childNodes[2];

                //only append this param if its checkbox is checked
                if( (bIgnoreCheck == true) || oCheck.getAttribute("checked") == "true" )
                {
                    //don't append this, if its label is empty
                    if( oLabel.value != "" )
                    {
                        var sLabel = String(oLabel.value);
                        var sValue = "";

                        // strip [var] or [url] at the end of the label
                        var sStrippedLabel = sLabel.replace( /\[(var|url)\]$/, "" );

                        // check if aParams already contains an entry for sStrippedLabel
                        var uknSearch = searchParameterArray( sStrippedLabel, aParams );
                        if( typeof(uknSearch) != "boolean" ) {
                            sValue = uknSearch[1];
                        }

                        var bUrlVar = false;

                        // check wether sLabel contains [url] or [var] at the end
                        if( rxUrl.test(sLabel) ) {
                            sValue = oValue.value + "?";
                            bUrlVar = true;
                        }
                        else if( rxVar.test(sLabel) ) {
                            sValue = sValue + "&" + oValue.value;
                            bUrlVar = true;
                        }
                        else {
                            sValue = oValue.value;
                            bUrlVar = false;
                        }

                        if( bUrlVar == false || typeof(uknSearch) == "boolean" ) {
                            aParams.push( new Array(sStrippedLabel, sValue, oCheck.checked) );
                        } else {
                            uknSearch[1] = sValue;
                        }

                        this.debugDump( "param: [" + sStrippedLabel + "] = " + sValue );
                    }
                } else if( (bIgnoreCheck == false) && (this.bKeepUncheckedParams == true) ) {

                    if( this.aUncheckedParameters == null ) {
                        this.aUncheckedParameters = new Array();

                        this.aUncheckedParameters[id] = new Array();
                    }

                    if( oLabel.value != "" ) {
                        this.aUncheckedParameters[id].push( new Array(oLabel.value,oValue.value,oCheck.checked) );
                    }
                }
            }
        }

        return aParams;
    },

    getDataString: function(id, bIgnoreCheck)
    {
        this.debugDump( "getDataString start...." );

        if( arguments.length < 2 ) {
            var bIgnoreCheck = false;
        }

        var sParams = "";
        var aParams = this.getDataArray(id, bIgnoreCheck);

        for(var sKey in aParams) {
            var sLabel = aParams[sKey][0];
            var sValue = aParams[sKey][1];

            if( this.bUrlDeEncodeUrl ) {
                sValue = urlencode( sValue );
                sLabel = urlencode( sLabel );
            }

            sParams = sParams + "&" + sLabel + "=" + sValue;
        }

        if( sParams != "" ) {
            //delete the first &
            sParams = sParams.substr(1, sParams.length - 1);
        }

        this.debugDump("getDataString ("+id+"): "+sParams);

        return sParams;
    },

    redirectTab: function(bNewTab)
    {
        //var currentbrowser = this.oTabBrowser.selectedBrowser;
        //currentbrowser.loadURI(sNewUrl);

        var getData      = this.getDataString("gridrows-get");
        var postData     = this.getDataString("gridrows-post");
        var sUrl         = this.theWindow.document.getElementById("url_wo_params").value;
        var referringUri = this.theWindow.document.getElementById("url_referer").value;

        var postDataStream;
        if (postData != "")
        {
            var postDataHeaders = "Content-Type: application/x-www-form-urlencoded\r\nContent-Length: "+postData.length+"\r\n\r\n";
            postDataStream      = Components.classes["@mozilla.org/io/string-input-stream;1"].createInstance(Components.interfaces.nsIStringInputStream);

            postDataStream.setData(postDataHeaders + postData, -1);
        }
        else
        {
            postDataStream = null;
        }

        var uri   = Components.classes["@mozilla.org/network/standard-url;1"].createInstance(Components.interfaces.nsIStandardURL);
        uri       = uri.QueryInterface(Components.interfaces.nsIURL);
        uri.spec  = sUrl;
        uri.query = getData;

        if (referringUri=="")
        {
            referringUri = null;
        }
        else
        {
            var referringUriStr = referringUri;
            referringUri = Components.classes["@mozilla.org/network/simple-uri;1"].createInstance(Components.interfaces.nsIURI);
            try
            {
                referringUri.spec = referringUriStr;
            }
            catch (e)
            {
                alert('Referring URI \''+referringUriStr+'\' is not a valid URI.');
                throw e;
            }
        }

        if( bNewTab == true )
        {
            var newtab = this.oTabBrowser.addTab();
            this.oTabBrowser.selectedTab = newtab;
        }

        //oHeadersStream = Components.classes["@mozilla.org/io/string-input-stream;1"].createInstance(Components.interfaces.nsIStringInputStream);
        //oHeadersStream.setData("User-Agent: UrlParams v1.10.0\r\n", -1);

        var webNavigation = this.oTabBrowser.selectedBrowser.webNavigation;
        webNavigation.loadURI(uri.spec, webNavigation.LOAD_FLAGS_NONE, referringUri, postDataStream, null);
    },

    changeLoadEvent: function()
    {
        if( this.selectedBrowser != null )
        {
            this.debugDump("changeLoadEvent: removing load event listener.");

            try {
                this.selectedBrowser.removeProgressListener(this);
            } catch (e) {
                this.debugDump('error with removeProgressListener: ' + e);
            }
        }

        this.selectedBrowser = this.oTabBrowser.selectedBrowser;
        this.selectedBrowser.addProgressListener(this, Components.interfaces.nsIWebProgress.NOTIFY_LOCATION );
    },

    clearChecks: function(id)
    {
        //get rows object
        var oGridRows = this.theWindow.document.getElementById(id);

        //walk the childs and uncheck the checkbox (if the name col isn't empty
        for( var i = 0; i < oGridRows.childNodes.length; i++ )
        {
            var theRow = oGridRows.childNodes[i];

            if( theRow.childNodes[1].value != "" )
            {
                theRow.childNodes[0].setAttribute("checked", "false");

                if( theRow.childNodes[0].getAttribute("disabled") == "true" )
                {
                    theRow.childNodes[0].setAttribute("disabled", "false");
                }
            }
        }
    },

    switchGetPost: function() {
        var aGetData  = this.getDataArray("gridrows-get" , true);
        var aPostData = this.getDataArray("gridrows-post", true);

        this.removeGridChilds();

        var iPosts = this.fillBoxFromArray("gridrows-post", aGetData );
        var iGets  = this.fillBoxFromArray("gridrows-get" , aPostData);

        this.addEmptyRows();

        this.theWindow.document.getElementById("groupboxGetLabel").value  = "GET ("  + iGets  + ")";
        this.theWindow.document.getElementById("groupboxPostLabel").value = "POST (" + iPosts + ")";
    },

    transferPost2Get: function() {
        var aGetData  = this.getDataArray("gridrows-get" , true);
        var aPostData = this.getDataArray("gridrows-post", true);

        this.removeGridChilds();

        var iGets = this.fillBoxFromArray("gridrows-get", aGetData );
            iGets = iGets + this.fillBoxFromArray("gridrows-get" , aPostData);

        this.addEmptyRows();

        this.theWindow.document.getElementById("groupboxGetLabel").value  = "GET ("  + iGets  + ")";
        this.theWindow.document.getElementById("groupboxPostLabel").value = "POST (0)";
    },

    transferGet2Post: function() {
        var aGetData  = this.getDataArray("gridrows-get" , true);
        var aPostData = this.getDataArray("gridrows-post", true);

        this.removeGridChilds();

        var iPosts = this.fillBoxFromArray("gridrows-post", aGetData );
            iPosts = iPosts + this.fillBoxFromArray("gridrows-post", aPostData);

        this.addEmptyRows();

        this.theWindow.document.getElementById("groupboxGetLabel").value  = "GET (0)";
        this.theWindow.document.getElementById("groupboxPostLabel").value = "POST (" + iPosts + ")";
    },

    removeDuplicates: function() {
        var aGetData  = removeDuplsFromArray(this.getDataArray("gridrows-get" , true));
        var aPostData = removeDuplsFromArray(this.getDataArray("gridrows-post", true));

        this.removeGridChilds();

        var iPosts = this.fillBoxFromArray("gridrows-post", aPostData );
        var iGets  = this.fillBoxFromArray("gridrows-get" , aGetData);

        this.addEmptyRows();

        this.theWindow.document.getElementById("groupboxGetLabel").value  = "GET ("  + iGets  + ")";
        this.theWindow.document.getElementById("groupboxPostLabel").value = "POST (" + iPosts + ")";
    },

    updateFormHistory: function() {

        this.debugDump("updateFormHistory: updating...");

        if( this.sAutocomplete == 'off' ) {
            this.debugDump("updateFormHistory: autocomplete is not enabled. do nothing.");
            return;
        }

        // well, @mozilla.org/satchel/form-history;1 does not exist in mozilla, so catch this
        // and return on error
        var oFormHistory = null;
        try {
            oFormHistory = Components.classes["@mozilla.org/satchel/form-history;1"].getService(Components.interfaces.nsIFormHistory);
        } catch (e) {
            return;
        }

        var aObjects = this.theWindow.document.getElementsByTagName("textbox");
        this.debugDump("updateFormHistory: found " + aObjects.length + " textboxes. movin' on...");

        for( var i=0; i < aObjects.length; i++ )
        {
            var oElement = aObjects[i];
            this.debugDump("updateFormHistory: Element has type: " + oElement.getAttribute("type"));

            if( oElement.getAttribute("type") == "autocomplete" ) {
                var sValue = oElement.value;
                var sKey   = oElement.getAttribute("autocompletesearchparam");

                if( sValue != "" ) {
                    this.debugDump("adding \"" + sValue + "\" for field \"" + sKey + "\"");
                    oFormHistory.addEntry(sKey, sValue);
                }
            }
        }

        this.debugDump("updateFormHistory: done.");
    },

    exportData: function() {

        var oFileIO = new FileIO();

        oFileIO.defaultExtension = "xml";
        oFileIO.defaultString    = "UrlParamsExport.xml";

        oFileIO.appendFilter( "XML Files (*.xml)", "*.xml" );
        oFileIO.appendFilter( "All Files (*.*)", "*.*" );

        var bRes = oFileIO.pickFile( this.theWindow, "Choose file to export UrlParams data to", oFileIO.consts.modeSave );

        this.debugDump("bRes: " + bRes);
        this.debugDump("pickFileResult: " + oFileIO.pickFileResult);
        this.debugDump("returnOK: " + oFileIO.consts.returnOK);
        this.debugDump("returnReplace: " + oFileIO.consts.returnReplace);
        this.debugDump("returnCancel: " + oFileIO.consts.returnCancel);

        if( bRes ) {

            this.debugDump("Chosen File: " + oFileIO.getPickedFile.target);

            var aGetData   = this.getDataArray("gridrows-get");
            var aPostData  = this.getDataArray("gridrows-post");
            var sUrl       = this.theWindow.document.getElementById("url_wo_params").value;
            var sReferrer  = this.theWindow.document.getElementById("url_referer").value;

            var sFileData = new String();

            sFileData += "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";

            sFileData += "<request>\n";
            sFileData += "    <url>"+sUrl+"</url>\n";
            sFileData += "    <referrer>"+sReferrer+"</referrer>\n";

            sFileData += "    <gets>\n";
            for(var sKey in aGetData) {
                sFileData += "        <param name=\""+ aGetData[sKey][0] +"\">"+ aGetData[sKey][1] +"</param>\n";
            }
            sFileData += "    </gets>\n";

            sFileData += "    <posts>\n";
            for(var sKey in aPostData) {
                sFileData += "        <param name=\""+ aGetData[sKey][0] +"\">"+ aPostData[sKey][1] +"</param>\n";
            }
            sFileData += "    </posts>\n";


            sFileData += "</request>";

            this.debugDump( sFileData );

            oFileIO.writeFile( sFileData, 'UTF-8' );
        } else {
            this.debugDump( "you chose cancel" );
        }
    },

    ///////////////////////////////////////////////////////////////////////////
    /////// drag&drop handlers
    initDragDrop: function() {
        if( nsDragAndDrop ) {
            // have a look at the handleEvent function below

            this.theWindow.addEventListener("dragover", this, false);
            this.theWindow.addEventListener("dragdrop", this, false);
        }
    },

    onDragOver: function(event, flavour, session) {
        // well, do nothing, but has to be defined
    },

    onDrop: function(event, dropData, session) {
        this.debugDump("onDrop's data: " + dropData.data );
        this.debugDump("onDrop's flavour: " + dropData.flavour.contentType );

        var aDataParts = dropData.data.split("\n");
        var sUrl       = aDataParts[0];

        if( sUrl != "" ) {
            this.removeGridChilds();

            var oUrlBox      = this.theWindow.document.getElementById("url_wo_params");
            var oReferrerBox = this.theWindow.document.getElementById("url_referer");
            var boxGetLabel  = this.theWindow.document.getElementById("groupboxGetLabel");
            var boxPostLabel = this.theWindow.document.getElementById("groupboxPostLabel");

            // copy referrer from url textbox
            oReferrerBox.value = oUrlBox.value;

            var aUrlParts = sUrl.split("?");

            if( aUrlParts[0] != "" ) {
                oUrlBox.value = aUrlParts[0];
            }

            var iGets = 0;
            if( aUrlParts[1] ) {
                iGets = this.fillBox( "gridrows-get", aUrlParts[1] );
            }

            boxGetLabel.value  = "GET ("+iGets+")";
            boxPostLabel.value = "POST (0)";
            this.addEmptyRows();

        }
    },

    getSupportedFlavours: function() {
        var oFlavourSet = new FlavourSet();
        oFlavourSet.appendFlavour("text/x-moz-url");
        oFlavourSet.appendFlavour("text/unicode");
        return oFlavourSet;
    },

    ///////////////////////////////////////////////////////////////////////////
    /////// event handlers

    // main event dispatcher function
    handleEvent: function( event )
    {
        this.debugDump("handleEvent: " + event.type);

        if( !this.theWindow || !this.theWindow.document )
        {
            this.debugDump("aborting... theWindow or theWindow.document is empty");
            return
        }

        // dispatch the events
        switch( event.type )
        {
            case "load":     if( event.currentTarget == this.theWindow )
                             {
                                this.init(event);
                             }
                             else
                             {
                                this.onTabFocus(event);
                             }
                             break;

            case "unload":   this.onWindowClosing(event); break;

            case "click":    this.onClick(event); break;

            case "dblclick": this.onDblClick(event); break;

            case "select":   this.onTabFocus(event); break;

            case "command":  switch( event.currentTarget.id )
                             {
                                case "get-button":        this.updateFormHistory(); this.onGetInCurrentTab(event); break;
                                case "get-button-newtab": this.updateFormHistory(); this.onGetInNewTab(event);     break;
                                case "clear-checkboxes":  this.onClearCheckboxes(event); break;
                                case "get-url":           this.onGetURL(event);          break;
                             }
                             break;

            case "popuphidden":
            case "blur":     if( event.currentTarget ) {
                                 if( event.currentTarget.theLabel ) {
                                     this.onChangeBlurEditLabel(event);
                                 } else if( event.currentTarget.theValueTextbox ) {
                                     this.changeEmptyValueInput(event.currentTarget);
                                 }
                             }
                             break;

            case "change":   if( event.currentTarget && event.currentTarget.value) {
                                 if( event.currentTarget.theLabel ) {
                                     this.onChangeBlurEditLabel(event);
                                 } else {
                                     event.currentTarget.setAttribute("tooltiptext", event.currentTarget.value);
                                 }
                             }
                             break;

            case "dragover": nsDragAndDrop.dragOver(event, this);
                             break;

            case "dragdrop": nsDragAndDrop.drop(event, this);
                             break;
        }

        this.debugDump("handleEvent done");
    },

    onChangeBlurEditLabel: function(event) {
        var oTextbox    = event.currentTarget;
        var oRow        = oTextbox.parentNode;
        var oLabel      = oTextbox.theLabel;
        var oValueInput = oRow.childNodes[2];

        if( oTextbox.blurCount == 0 ) {
            oTextbox.blurCount = 1;
            return;
        }

        //////////////////////////////////////////////////////
        // just save new value, if its not an empty string
        if( oTextbox.value != "" ) {
            //////////////////////////////////////////////////////
            //save new value for the parameter name label
            oLabel.value = oTextbox.value;

            //////////////////////////////////////////////////////
            // change the value inputs autocompletesearchparam

            //saving value, cause this gets lost during convertion
            var sValue = oValueInput.value;

            //first, restore the textbox into a normal one, then
            //convert the textbox into an autocomplete textbox
            this.setAutocompleteTextboxAttributes(oValueInput, true);
            this.setAutocompleteTextboxAttributes(oValueInput, false, oLabel.value, false);

            oValueInput.value = sValue;
        }

        //////////////////////////////////////////////////////
        //remove event listeners and replace textbox with the label
        oTextbox.removeEventListener("change", this, true);
        oTextbox.removeEventListener("blur", this, true);

        oRow.replaceChild( oLabel, oTextbox );

        //give the value input box focus
        oValueInput.focus();
    },

    onGetInCurrentTab: function(event)
    {
        this.redirectTab(false);
    },

    onGetInNewTab: function()
    {
        this.redirectTab(true);
    },

    onClick: function(event) {
        if( event.explicitOriginalTarget ) {
            this.debugDump("onClick: caught click on id: " + event.explicitOriginalTarget.id);

            var bRightClick = false;
            if( event.button != 0 ) {
                bRightClick = true;

                //don't show the context menu
                event.preventDefault();
            }

            switch( event.explicitOriginalTarget.id ) {
                case "img-plusget":
                    this.addEmptyRow("gridrows-get", bRightClick);
                    break;
                case "img-pluspost":
                    this.addEmptyRow("gridrows-post", bRightClick);
                    break;
                default: break;
            }
        }
    },

    onDblClick: function(event)
    {
        if( event.explicitOriginalTarget ) {
            this.debugDump("onDblClick: caught dblclick on id: " + event.explicitOriginalTarget.id);

            switch( event.explicitOriginalTarget.tagName ) {
                case "rows":
                    this.debugDump("onDblClick: its a rows object...");

                    if( (event.explicitOriginalTarget.id == "gridrows-get") || (event.explicitOriginalTarget.id == "gridrows-post"))
                    {
                        this.debugDump("onDblClick: id is '" + event.explicitOriginalTarget.id + "', adding new empty line");

                        this.addEmptyRow(event.explicitOriginalTarget.id);
                    }
                    break;

                case "textbox":
                    this.debugDump("onDblClick: its a textbox...");

                    if( event.explicitOriginalTarget.getAttribute("type") == "autocomplete" ) {
                        this.debugDump("onDblClick: its an autocomplete checkbox. do the search");

                        var oTextbox = event.explicitOriginalTarget;

                        //start the search, this will implicitly open the popup (or not :) )
                        // but only start the search, if no search is currently in progress
                        if( oTextbox.controller && (oTextbox.controller.searchStatus != oTextbox.controller.STATUS_SEARCHING) ) {
                            oTextbox.controller.startSearch(event.explicitOriginalTarget.value);
                        }
                    }
                    break;

                case "label":
                    this.debugDump("onDblClick: its a label..");

                    var oLabel    = event.explicitOriginalTarget;

                    if( oLabel.parentNode && oLabel.parentNode.tagName == "row" ) {
                        this.debugDump("onDblClick: well, yeah, convert it");

                        var oRow      = oLabel.parentNode;

                        var oNewTextbox = this.theWindow.document.createElement("textbox");
                        oNewTextbox.setAttribute("type", "");
                        oNewTextbox.setAttribute("value", oLabel.value);
                        oNewTextbox.theLabel = oLabel;
                        oNewTextbox.blurCount = 0;

                        oRow.replaceChild( oNewTextbox, oLabel )

                        oNewTextbox.addEventListener("change", this, true);
                        oNewTextbox.addEventListener("blur", this, true);
                        oNewTextbox.focus();
                        oNewTextbox.select();
                    }

                    break;
                default:
                    this.debugDump("onDblClick: do nothing");
                    break;
            }
        }
    },

    onTabFocus: function(event)
    {
        this.debugDump("onTabFocus...");

        if( event )
        {
            this.debugDump("onTabFocus: event.type: " + event.type);

            if( event.requestingWindowURI )
            {
                this.debugDump("onTabFocus: event.requestingWindowURI: " + event.requestingWindowURI.spec);
            }
        }

        this.changeLoadEvent();

        this.removeGridChilds();
        this.fill();

        this.debugDump("...done");
    },

    onClearCheckboxes: function(event)
    {
        this.debugDump("onClearCheckboxes... ");

        this.clearChecks("gridrows-get");
        this.clearChecks("gridrows-post");

        this.debugDump("...done");
    },

    onGetURL: function(event)
    {
        this.debugDump("onGetURL... ");

        this.onTabFocus(null);

        this.debugDump("...done");
    },

    onWindowClosing: function(event)
    {
        this.debugDump("onWindowClosing... removing listeners");

        if( this.selectedBrowser != null )
        {
            this.debugDump("removing load event listener on this.selectedBrowser...");
            this.selectedBrowser.removeEventListener("load", this, true);
        }

        if( this.oTabBrowser && this.oTabBrowser.mTabBox )
        {
            this.debugDump("removing select event listener on this.oTabBrowser.mTabBox...");
            this.oTabBrowser.mTabBox.removeEventListener('select', this, true);
        }

        this.debugDump("...done");
    },

    ///////////////////////////////////
    // nsiWebProgressListener functions
    onLocationChange: function( webProgress, request, location ) {
        this.debugDump('onLocationChange ' + location.spec, 2);

        this.onTabFocus();

        return 0;
    },

    onProgressChange: function() {
    },
    onSecurityChange: function() {
    },
    onStateChange: function( webProgress, request, stateFlags, status ) {
        return 0;
    },
    onStatusChange: function( webProgress, request, status, message ) {
    },
    onLinkIconAvailable: function() {
    },

    QueryInterface : function(aIID) {
        if (aIID.equals(Components.interfaces.nsIWebProgressListener) ||
            aIID.equals(Components.interfaces.nsISupportsWeakReference) ||
            aIID.equals(Components.interfaces.nsISupports))
                return this;

        throw Components.results.NS_NOINTERFACE;
    },
}
